#!/usr/bin/env python3
'''\
Test the Multivariate Normal kl divergence
'''
######## Imports ########
import numpy as np
from scipy import stats
from gwalk.multivariate_normal import MultivariateNormal
from gwalk.multivariate_normal import params_of_offset_mu_cov
from gwalk.multivariate_normal import params_of_mu_cov
#from gwalk.multivariate_normal import multivariate_normal_marginal1d
#from gwalk.multivariate_normal import multivariate_normal_marginal2d
from gwalk.multivariate_normal import cov_of_std_cor
from gwalk.multivariate_normal import mu_cov_of_params
from gwalk.multivariate_normal import pdf as gwalk_norm
from gwalk.multivariate_normal import analytic_kl
from gwalk.multivariate_normal import analytic_kl_of_params
from scipy.stats import multivariate_normal as scipy_norm
from gp_api.utils import sample_hypercube
from basil_core.stats.distance import rel_entr as rel_entr_basil
from scipy.special import rel_entr as rel_entr_scipy
import time

######## Settings ########

N_RVS = int(1e6)
N_MAX = 3
HYPERCUBE_RES = 100
SEED = 0
rs = np.random.RandomState(SEED)
LIMITS = np.asarray([
                     [-7., 7.],
                     [-7., 7.],
                    ])

######## Tools ######## 
def make_gaussians():
    ''' Create 3 MV objects, one offset, and one different'''
    # Create shared limits
    #Identify the dimensionality of our Gaussians
    ndim = len(LIMITS)

    # lnl_offset is not required for KL test
    norm = 0.

    # Initialize mu parameters
    mu1 = np.asarray([0.5,0.5])
    delta_mu = np.asarray([1.,0.])
    mu2 = mu1 + delta_mu
    mu3 = np.asarray([0.3, 0.45])

    # Initialize sig parameters
    sig1 = np.ones(ndim)
    sig3 = np.ones(ndim)*1.358

    # Pick scale parameters
    scale1 = np.copy(sig1)
    scale3 = sig3 + 0.285

    # Pick standard correlation parameters
    rho1 = np.asarray([
                      [1., 0.,],
                      [0., 1.,],
                     ])
    rho_ij = 0.6
    rho3 = np.asarray([
                       [1., rho_ij,],
                       [rho_ij, 1.,],
                      ])

    # Get Covariance matrices
    Sig1 = cov_of_std_cor(sig1, rho1)[0]
    Sig3 = cov_of_std_cor(sig3, rho3)[0]


    # Identify parameters
    X1 = params_of_mu_cov(mu1, Sig1).flatten()
    # Note we do use Sig1 again for X2
    X2 = params_of_mu_cov(mu2, Sig1).flatten()
    X3 = params_of_mu_cov(mu3, Sig3).flatten()

    # Create scaled Parameterizations
    X1s = np.copy(X1)
    X1s[1:ndim+1] /= scale1
    X1s[ndim+1:2*ndim+1] /= scale1
    X2s = np.copy(X2)
    X2s[1:ndim+1] /= scale1
    X2s[ndim+1:2*ndim+1] /= scale1
    X3s = np.copy(X3)
    X3s[1:ndim+1] /= scale3
    X3s[ndim+1:2*ndim+1] /= scale3
    
    # Create multivariate normal objects
    MV1 = MultivariateNormal(X1, scale=scale1, seed=SEED, limits=LIMITS)
    # Note we use scale1 for MV2
    MV2 = MultivariateNormal(X2, scale=scale1, seed=SEED, limits=LIMITS)
    MV3 = MultivariateNormal(X3, scale=scale3, seed=SEED, limits=LIMITS)

    # Assign the true parameterizations to the MV objects
    MV1.params = X1
    MV2.params = X2
    MV3.params = X3

    return MV1, MV2, MV3

def generate_samples(MV,n_rvs):
    '''Generate samples from the Gaussian'''
    samples = MV.sample_normal(n_rvs)
    return samples

def build_hypercube(MV, hypercube_res):
    '''Build a grid of normalized likelihood evaluations'''
    #samples = MV.mu_hypercube(res=hypercube_res)
    samples = sample_hypercube(LIMITS,res=hypercube_res)
    lnL = MV.likelihood(samples, log_scale=True).flatten()
    L = np.exp(lnL)
    L_sum = np.sum(L)
    L /= L_sum
    lnL_offset = np.log(L_sum)
    lnL -= lnL_offset
    return samples, L, lnL

######## main ########

def main():
    ## Build MultivariateNormal
    MV1, MV2, MV3 = make_gaussians()
    # Build likelihood grid
    Y_grid, L_true, lnL_true = build_hypercube(MV1, HYPERCUBE_RES)
    # Build offset grid
    Y_grid, L_false, lnL_false = build_hypercube(MV2, HYPERCUBE_RES)


    # Test relative entropy
    print("Testing relative entropy kl:")
    t0 = time.time()
    kl_scipy = np.sum(rel_entr_scipy(L_true, L_false))
    t1 = time.time()
    kl_basil = np.sum(rel_entr_basil(L_true, lnP=lnL_true, lnQ=lnL_false[None,:], normQ=False))
    t2 = time.time()
    print("Scipy time: %f; kl: %f"%(t1-t0,kl_scipy))
    print("basil time: %f; kl: %f"%(t2-t1, kl_basil))
    assert np.allclose(kl_basil, kl_scipy)
    print("  pass!")

    # Extract parameters
    X1 = MV1.params
    mu1, Sig1 = mu_cov_of_params(X1)
    X2 = MV2.params
    mu2, Sig2 = mu_cov_of_params(X2)
    X3 = MV3.params
    mu3, Sig3 = mu_cov_of_params(X3)

    # Test analytic kl divergence for mu offset with same scale
    print("Testing analytic kl:")
    kl12_analytic = analytic_kl(mu1, Sig1, mu2, Sig2)
    kl12_analytic_params = analytic_kl_of_params(X1, X2)
    kl12_analytic_obj = MV1.analytic_kl(X2)
    assert np.allclose(kl12_analytic, kl_scipy)
    assert np.allclose(kl12_analytic_params, kl_scipy)
    assert np.allclose(kl12_analytic_obj, kl_scipy)
    print("  pass!")

    # Test analytic kl for Gaussians with different scales and covariance
    print("Testing analytic kl with scaling:")
    kl13_analytic = analytic_kl(mu1, Sig1, mu3, Sig3, scale1=MV1.scale, scale2=MV3.scale)
    kl13p = analytic_kl_of_params(X1, X3, scale1=MV1.scale, scale2=MV3.scale)
    kl13s = analytic_kl_of_params(X1, X3, scale1=MV1.scale, scale2=MV3.scale)
    kl13obj1 = MV1.analytic_kl(X3, scale2=MV3.scale)
    kl13objobj = MV1.analytic_kl(MV3)
    assert np.allclose(kl13_analytic, kl13p)
    assert np.allclose(kl13_analytic, kl13s)
    assert np.allclose(kl13_analytic, kl13obj1)
    assert np.allclose(kl13_analytic, kl13objobj)
    print("  pass!")

    return

######## Execution ########
if __name__ == "__main__":
    main()



''' Prior functions for gwalk
'''
######## Imports ########

from gwalk.catalog.prior.CIP_prior_functions import mass_area, eta_norm_factor, mc_prior, eta_prior
from gwalk.catalog.prior.callister_prior import chi_effective_prior_of_aligned_spins as chi_eff_aligned_prior
from gwalk.catalog.coordinates import q_of_mc_eta

######## Functions ########

def prior_mc_eta(
                 mc, eta,
                 mc_min=0., mc_max=100.,
                 eta_min=0., eta_max=0.24999999,
                 spin_max=1.,
                ):
    '''Generate prior values for mc and eta samples

    Parameters
    ----------
    mc: array like, shape = (npts)
        Input mc values
    eta: array like, shape = (npts)
        Input eta values
    mc_min: float
        Input minimum value for mc
    mc_max: float
        Input maximum value for mc
    eta_min: float
        Input minimum value for eta
    eta_max: float
        Input maximum value for eta
    spin_max: float, optional
        Input maximum allowed spin
    '''
    import numpy as np
    # Generate the eta_norm_factor
    # TODO investigate this line
    norm_factor = eta_norm_factor(eta_min, eta_max)
    # Generate the mass area
    A = mass_area(mc_min, mc_max, eta_min, eta_max)
    # Generate the mc prior weights
    mc_weights = mc_prior(mc, mc_min, mc_max)
    # Generate the eta prior weights
    eta_weights = eta_prior(eta, norm_factor = norm_factor)
    # Find the mass prior
    prior = A * mc_weights * eta_weights

    # Positive semidefinite
    keep = prior > 0
    prior[~keep] = 0.0
    pmin, pmax = np.min(prior[keep]), np.max(prior[keep])
    if pmin > 1.:
        prior[keep] /= pmin
    elif pmax < 1.:
        prior[keep] /= pmax
    return prior

def prior_mc_eta_chieff(
                        mc, eta, chieff,
                        mc_min=0., mc_max=100.,
                        eta_min=0., eta_max=0.24999999,
                        spin_max=1.,
                       ):
    '''Generate uninformative prior values for mc,eta, and chi effective

    Parameters
    ----------
    mc: array like, shape = (npts)
        Input mc values
    eta: array like, shape = (npts)
        Input eta values
    chieff: array like, shape = (npts)
        Input chi effective values
    mc_min: float
        Input minimum value for mc
    mc_max: float
        Input maximum value for mc
    eta_min: float
        Input minimum value for eta
    eta_max: float
        Input maximum value for eta
    spin_max: float, optional
        Input maximum allowed spin
    '''
    import numpy as np
    # Find the mass prior
    mass_prior = prior_mc_eta(
                              mc, eta,
                              mc_min=mc_min, mc_max=mc_max,
                              eta_min=eta_min, eta_max=eta_max,
                             )
            

    # Find mass ratios
    q = q_of_mc_eta(mc, eta)
    # Find the chi effective prior
    chieff_prior = chi_eff_aligned_prior(q, spin_max, chieff)

    # Combine prior
    prior = mass_prior * chieff_prior
    # Positive semidefinite
    prior[prior < 0] = 0.0

    return prior

def prior_dist(lum_dist):
    '''luminosity distance prior
    Parameters
    ----------
    lum_dist: array like, shape = (npts)
        Input luminosity distance values
    '''
    import numpy as np
    p = np.power(lum_dist, 2.)
    pmin, pmax = np.min(p), np.max(p)
    if pmin > 1.:
        p /= pmin
    elif pmax < 1.:
        p /= pmax
    return p

def prior_full_spin(chi1x, chi2x, chi1y, chi2y, chi1z, chi2z):
    '''6 dimensional spin component prior

    Parameters
    ----------
    chi1x: array like, shape = (npts)
        Input spin values
    chi2x: array like, shape = (npts)
        Input spin values
    chi1y: array like, shape = (npts)
        Input spin values
    chi2y: array like, shape = (npts)
        Input spin values
    chi1z: array like, shape = (npts)
        Input spin values
    chi2z: array like, shape = (npts)
        Input spin values
    '''
    import numpy as np
    chi1sq = chi1x**2 + chi1y**2 + chi1z**2
    chi2sq = chi2x**2 + chi2y**2 + chi2z**2
    p = np.power(chi1sq*chi2sq,-1.)
    return p



#!/usr/env/bin python3

######## Setup ########
import numpy as np
import time

######## Globals ########
MAX_NDIM = 100
NDIM = 3
NPTS = int(1e4)
NGAUSS = int(1e2)
SEED = 0
RS = np.random.RandomState(SEED)

def test_properties(ndim):
    ''' Test the pdf evaluation with scale == 1'''
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.multivariate_normal import mu_cov_of_params
    from gwalk.multivariate_normal import std_of_cov, cor_of_std_cov
    from gwalk.multivariate_normal.object import MultivariateNormal

    #### Setup ####
    RS.seed(SEED)
    X0 = random_gauss_params(1, ndim, rs=RS)
    X0[:,ndim + 1] = 1e-10
    X0[:,ndim + 2] = 1e+10
    X0[:,2*ndim + 1:] *= 0.5
    X0 = X0.flatten()
    offset = X0[0]
    mu, cov = mu_cov_of_params(X0)
    mu, cov = mu.flatten(), cov.reshape((ndim,ndim))
    std = std_of_cov(cov)
    cor = cor_of_std_cov(std, cov)

    #### Initialize object ####
    print("Testing MV initialization:")
    MV = MultivariateNormal(X0)
    assert MV.ndim == ndim
    assert np.allclose(MV.params, X0)
    assert np.allclose(MV.offset, X0[0])
    assert np.allclose(MV.mu, mu)
    assert np.allclose(MV.std, std)
    assert np.allclose(MV.cor, cor)
    assert np.allclose(MV.cov, cov)
    print("  pass!")

def test_limits(ndim):
    ''' Test limit setting and reading '''
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.multivariate_normal import mu_cov_of_params
    from gwalk.multivariate_normal import std_of_cov, cor_of_std_cov
    from gwalk.multivariate_normal.object import MultivariateNormal

    #### Setup ####
    RS.seed(SEED)
    X0 = random_gauss_params(1, ndim, rs=RS)
    X0[:,2*ndim + 1:] *= 0.5
    X0 = X0.flatten()
    offset = X0[0]
    mu, cov = mu_cov_of_params(X0)
    mu, cov = mu.flatten(), cov.reshape((ndim,ndim))
    std = std_of_cov(cov)
    cor = cor_of_std_cov(std, cov)
    ## Set limits ##
    limits = np.zeros((ndim, 2))
    limits[:,1] = 1.

    #### Test limits ####
    print("Testing object limits:")
    MV0 = MultivariateNormal(X0)
    MV0.limits = limits
    MV1 = MultivariateNormal(X0, limits=limits)
    assert np.allclose(MV1.limits, limits)
    assert np.allclose(MV1.limits, MV0.limits)
    assert np.allclose(MV1.plimits, MV0.plimits)
    print("  pass!")

def test_likelihood(ndim, ngauss, npts):
    '''Test likelihood evaluation'''
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.multivariate_normal import mu_cov_of_params
    from gwalk.multivariate_normal import std_of_cov, cor_of_std_cov
    from gwalk.multivariate_normal.object import MultivariateNormal
    from gwalk.multivariate_normal import pdf
    from gwalk.multivariate_normal import std_of_params
    from gwalk.multivariate_normal import params_reduce

    #### Setup ####
    RS.seed(SEED)
    X0 = random_gauss_params(1, ndim, rs=RS)
    X0[:,2*ndim + 1:] *= 0.5
    X0 = X0.flatten()
    Xn = random_gauss_params(ngauss, ndim, rs=RS)
    Xn[:,2*ndim + 1:] *= 0.5
    mu0, cov0 = mu_cov_of_params(X0)
    mun, covn = mu_cov_of_params(Xn)
    indices = [0,1]

    ## Set limits ##
    limits = np.zeros((ndim, 2))
    limits[:,1] = 1.
    MV0 = MultivariateNormal(X0, limits=limits)

    ## Generate samples ##
    Y0 = RS.uniform(size=npts*ndim).reshape((npts,ndim))
    L0 = pdf(mu0, cov0, Y0, scale=MV0.scale, log_scale=False)
    lnL0 = pdf(mu0, cov0, Y0, scale=MV0.scale, log_scale=True)
    Ln = pdf(mun, covn, Y0, scale=MV0.scale, log_scale=False)
    lnLn = pdf(mun, covn, Y0, scale=MV0.scale, log_scale=True)

    #### Test likelihood ####
    print("Testing likelihood:")
    L1 = MV0.likelihood(Y0,offset=0.)
    assert np.allclose(L0, L1)
    print("  pass!")

    print("Testing offset:")
    L2 = MV0.likelihood(Y0)
    assert np.allclose(L2, L0*np.exp(MV0.offset))
    print("  pass!")

    print("Testing log likelihood:")
    lnL1 = MV0.likelihood(Y0, offset=0, log_scale=True)
    assert np.allclose(lnL1, lnL0)
    print("  pass!")
    
    print("Testing log offset:")
    lnL1 = MV0.likelihood(Y0, log_scale=True)
    assert np.allclose(lnL1, lnL0 + MV0.offset)
    print("  pass!")

    print("Testing Xn likelihood:")
    Ln1 = MV0.likelihood(Y0, X=Xn, offset=0.)
    assert np.allclose(Ln1, Ln)
    print("  pass!")

    print("Testing Xn offset:")
    Ln1 = MV0.likelihood(Y0, X=Xn, offset=MV0.offset)
    assert np.allclose(Ln1, Ln*np.exp(MV0.offset))
    print("  pass!")

    print("Testing Xn log likelihood:")
    lnLn1 = MV0.likelihood(Y0, X=Xn, offset=0., log_scale=True)
    assert np.allclose(lnLn1, lnLn)
    print("  pass!")

    print("Testing Xn log offset:")
    lnLn1 = MV0.likelihood(Y0, X=Xn, log_scale=True, offset=MV0.offset)
    assert np.allclose(lnLn1, lnLn + MV0.offset)
    print("  pass!")

    #### Test scale ####
    print("Testing scale:")
    scale = std_of_params(Xn)
    Lns = pdf(mun, covn, Y0, scale=scale) * np.exp(MV0.offset)
    Lns1 = MV0.likelihood(Y0, X=Xn, scale=scale, offset=MV0.offset)
    assert np.allclose(Lns, Lns1)
    print("  pass!")

    #### Test limits ####
    print("Testing limits:")
    limits_1 = np.copy(limits)
    limits_1[:,1] = 0.5
    keep = np.ones((npts,),dtype=bool)
    for i in range(ndim):
        keep &= Y0[:,i] >= limits_1[i,0]
        keep &= Y0[:,i] <= limits_1[i,1]
    Lk = MV0.likelihood(Y0, limits=limits_1, offset=0.)
    assert np.allclose(Lk[keep], L0[keep])
    assert np.allclose(Lk[~keep], 0.)
    print("  pass!")

    #### Test indices ####
    print("Testing indices (%s):"%(str(indices)))
    X2 = params_reduce(X0, indices)
    mu2, cov2 = mu_cov_of_params(X2)
    mu2, cov2 = mu2.flatten(), cov2.reshape((len(indices),len(indices)))
    Y2 = Y0[:,indices]
    Lind0 = pdf(mu2, cov2, Y2, scale=MV0.scale[indices])
    Lind2 = MV0.likelihood(Y2, indices=indices, offset=0.)
    assert np.allclose(Lind0, Lind2)
    print("  pass!")

    print("Testing indices [0]:")
    #TODO: this works, but it shouldn't require this much care
    X3 = params_reduce(X0, [0])
    mu3, cov3 = mu_cov_of_params(X3)
    mu3, cov3 = mu3.flatten(), cov3.reshape((1,1))
    Y3 = Y0[:,0].reshape((npts,1))
    Li0 = pdf(mu3, cov3, Y3, scale=np.asarray([MV0.scale[0]]).reshape((1,)))
    Li3 = MV0.likelihood(Y3, indices=[0], offset=0.)
    assert np.allclose(Li0, Li3)
    print("  pass!")

def test_sample(ndim, npts):
    '''Test sample_normal'''
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.multivariate_normal import mu_cov_of_params
    from gwalk.multivariate_normal.object import MultivariateNormal

    #### Setup ####
    RS.seed(SEED)
    X0 = random_gauss_params(1, ndim, rs=RS)
    X0[:,2*ndim + 1:] *= 0.5
    X0 = X0.flatten()

    ## Set limits ##
    limits = np.zeros((ndim, 2))
    limits[:,1] = 1.

    ## Initialize MV ##
    MV = MultivariateNormal(X0, limits=limits)

    ## Test sampling ##
    print("Testing sampling:")
    Y0 = MV.sample_normal(size=npts)
    # Check shape of sample
    assert Y0.shape == (npts, ndim)
    # Check that all samples are inside limits
    for i in range(ndim):
        assert all(Y0[:,i] >= limits[i,0])
        assert all(Y0[:,i] <= limits[i,1])

    # Mess with the variance
    MV.std = MV.std / 40
    Y0 = MV.sample_normal(size=npts)
    # Check shape of sample
    assert Y0.shape == (npts, ndim)
    # Check that all samples are inside limits
    for i in range(ndim):
        assert all(Y0[:,i] >= limits[i,0])
        assert all(Y0[:,i] <= limits[i,1])
    frac_err = (MV.mu - np.mean(Y0, axis=0))/MV.mu
    # I hope 10000 samples don't have 5 percent error
    assert np.max(np.abs(frac_err)) < 0.05

    # Mess with the variance
    MV.std = MV.std * 4
    Y0 = MV.sample_normal(size=npts)
    frac_err = (MV.mu - np.mean(Y0, axis=0))/MV.mu
    # Check shape of sample
    assert Y0.shape == (npts, ndim)
    # Check that all samples are inside limits
    for i in range(ndim):
        assert all(Y0[:,i] >= limits[i,0])
        assert all(Y0[:,i] <= limits[i,1])

    # Check shape of sample
    assert Y0.shape == (npts, ndim)
    # Check that all samples are inside limits
    for i in range(ndim):
        assert all(Y0[:,i] >= limits[i,0])
        assert all(Y0[:,i] <= limits[i,1])
    print("  pass!")

def test_simple_fit(ndim, npts):
    '''Test simple_fit'''
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.multivariate_normal import mu_cov_of_params
    from gwalk.multivariate_normal.object import MultivariateNormal

    #### Setup ####
    RS.seed(SEED)
    X0 = random_gauss_params(1, ndim, rs=RS)
    X0[:,2*ndim + 1:] *= 0.5
    mu, cov = mu_cov_of_params(X0)
    X0 = X0.flatten()

    ## Set limits ##
    limits = np.zeros((ndim, 2))
    limits[:,0] = -10.
    limits[:,1] = 10.

    ## Initialize MV ##
    MV = MultivariateNormal(X0, limits=limits)

    ## Draw samples ##
    Y0 = MV.sample_normal(npts)

    ## Fit samples ##
    print("Testing fit_simple:")
    X1 = MV.fit_simple(Y0, assign=False)

    # Analytic kl divergence
    kl = MV.analytic_kl(X0, X1)
    assert kl < 1e-3
    print("  pass!")

def test_integral(ndim, ngauss):
    ''' Test the analytic enclosed integral'''
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.multivariate_normal import mu_cov_of_params
    from gwalk.multivariate_normal.object import MultivariateNormal
    from gwalk.multivariate_normal.utils import analytic_enclosed_integral

    #### Setup ####
    RS.seed(SEED)
    X0 = random_gauss_params(1, ndim, rs=RS)
    X0[:,2*ndim + 1:] *= 0.5
    mu, cov = mu_cov_of_params(X0)
    X0 = X0.flatten()
    Xn = random_gauss_params(ngauss, ndim, rs=RS)
    Xn[:,2*ndim + 1:] *= 0.5

    ## Set limits ##
    limits = np.zeros((ndim, 2))
    limits[:,0] = 0.
    limits[:,1] = 1.

    ## Initialize MV ##
    MV = MultivariateNormal(X0, limits=limits)

    ## Test integral ##
    print("Testing analytic integral:")
    # Test the function version of the enclosed integral
    integral0 = analytic_enclosed_integral(X0, limits, np.ones(MV.ndim))
    # Test the the object version
    integral1 = MV.analytic_enclosed_integral()
    # Check that they are the same
    assert np.allclose(integral0, integral1)
    # Test the assign feature and offset setting
    MV.analytic_enclosed_integral(assign=True)
    integral2 = np.exp(-MV.offset)
    assert np.allclose(integral0, integral2)

    # Test the function version with many inputs 
    integraln = analytic_enclosed_integral(Xn, limits)
    # Test the object version with different parameter inputs
    integraln1 = MV.analytic_enclosed_integral(X=Xn)
    # Assert that they are the same
    assert np.allclose(integraln, integraln1)
    print("  pass!")

def test_marginal(ndim, ngauss, npts):
    ''' Test the analytic enclosed integral'''
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.multivariate_normal import mu_cov_of_params
    from gwalk.multivariate_normal.object import MultivariateNormal
    from gwalk.multivariate_normal.utils import analytic_enclosed_integral

    #### Setup ####
    RS.seed(SEED)
    X0 = random_gauss_params(1, ndim, rs=RS)
    mu, cov = mu_cov_of_params(X0)
    X0 = X0.flatten()
    indices = [0,1]
    assert ndim > 2

    ## Get some samples ##
    Y0 = RS.uniform(size=npts*ndim).reshape((npts,ndim))
    Y1 = Y0[:,0].reshape((npts,1))
    Y2 = Y0[:,indices]


    ## Set limits ##
    limits = np.zeros((ndim, 2))
    limits[:,0] = 0.
    limits[:,1] = 1.

    ## Initialize MV ##
    MV = MultivariateNormal(X0, limits=limits)
    L01 = MV.likelihood(Y1, indices=[0], offset=0.)
    L02 = MV.likelihood(Y2, indices=indices, offset=0.)

    ## Test child ##
    print("Testing child (1-D):")
    child = MV.get_marginal([0])
    L1 = child.likelihood(Y1, offset=0.)
    assert np.allclose(L01, L1)
    print("  pass!")

    print("Testing child (%d-D):"%(len(indices)))
    child = MV.get_marginal(indices)
    L2 = child.likelihood(Y2, offset=0.)
    assert np.allclose(L02, L2)
    print("  pass!")

def test_serialization(ndim):
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.multivariate_normal import mu_cov_of_params
    from gwalk.multivariate_normal.object import MultivariateNormal
    from gwalk.multivariate_normal.utils import analytic_enclosed_integral

    #### Setup ####
    RS.seed(SEED)
    X0 = random_gauss_params(1, ndim, rs=RS)
    mu, cov = mu_cov_of_params(X0)
    X0 = X0.flatten()

    ## Set limits ##
    limits = np.zeros((ndim, 2))
    limits[:,0] = 0.
    limits[:,1] = 1.

    ## Initialize MV ##
    MV = MultivariateNormal(X0, limits=limits)
    fname = "test_object.hdf5"
    label = "my_label"
    attrs = {"butter": "apple"}

    print("Testing serialization: ")
    MV.save(fname, label, attrs=attrs)
    MVnew = MV.load(fname, label)
    assert MVnew == MV
    print("  pass!")

def test_from_properties(ndim):
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.multivariate_normal import mu_cov_of_params
    from gwalk.multivariate_normal import offset_of_params
    from gwalk.multivariate_normal.object import MultivariateNormal
    from gwalk.multivariate_normal.utils import analytic_enclosed_integral

    #### Setup ####
    RS.seed(SEED)
    X0 = random_gauss_params(1, ndim, rs=RS)
    mu, cov = mu_cov_of_params(X0)
    X0 = X0.flatten()

    ## Set limits ##
    limits = np.zeros((ndim, 2))
    limits[:,0] = 0.
    limits[:,1] = 1.

    ## Initialize MV ##
    MV0 = MultivariateNormal(X0, limits=limits)

    print("Testing from_properties:")
    MV1 = MultivariateNormal.from_properties(mu, cov=cov, offset=offset_of_params(X0), limits=limits)
    assert MV1 == MV0
    print("  pass!")


######## Main ########
def main():
    ngauss = NGAUSS
    ndim = NDIM
    npts = NPTS
    test_properties(ndim)
    test_limits(ndim)
    test_likelihood(ndim, ngauss, npts)
    test_sample(ndim, npts)
    test_simple_fit(ndim, npts)
    test_integral(ndim, ngauss)
    test_marginal(ndim, ngauss, npts)
    test_serialization(ndim)
    test_from_properties(ndim)
    return

######## Execution ########
if __name__ == "__main__":
    main()



#!/usr/bin/env python3
'''\
generate plots on a living axis
'''
from .percentile_levels import percentile_levels

######## Axis Plot Methods ########
#### 1 Dimensional Plots ####

def histogram_1d(
                 ax,
                 x,
                 w = None,
                 xmin = None,
                 xmax = None,
                 log_scale = False,
                 log_offset = 1e-4,
                 bins = 100,
                 **kwargs
                ):
    '''\
    Quickly make a histogram of some samples
    '''
    import numpy as np
    
    # Update limits
    if xmin is None:
        xmin = np.min(x)
    if xmax is None:
        xmax = np.max(x)

    # Make histogram
    rho, xedge = \
            np.histogram(
                         x,
                         weights = w,
                         range = [xmin, xmax],
                         bins = bins,
                        )

    #rho = rho / np.sum(rho)
    dx = (xmax - xmin)/bins
    rho = rho / np.sum(rho*dx)

    # log scale
    if log_scale:
        # Check min val
        min_val = np.min(rho[rho > 0.])
        # Implement log scale
        rho = np.log(rho + log_offset*min_val)

    # Show data
    ax.plot(np.linspace(xmin, xmax, bins), rho, **kwargs)
    ax.set_xlim([xmin, xmax])
    return ax




#### 2 Dimensional Plots ####

def scatter_2d(
               ax,
               x,
               y,
               pt = 5,
               w = None,
               xmin = None,
               xmax = None,
               ymin = None,
               ymax = None,
               **kwargs
              ):
    '''\
    Quickly plot samples in a scatter plot
    '''
    import numpy as np
    
    # Update limits
    if xmin is None:
        xmin = np.min(x)
    if xmax is None:
        xmax = np.max(x)
    if ymin is None:
        ymin = np.min(y)
    if ymax is None:
        ymax = np.max(y)

    if w is None:
        colors = 'r'
    else:
        colors = w
    
    # plot samples
    ax.scatter(x, y, c = colors, s = pt, edgecolors = 'k', linewidths = 0.2)
    ax.set_xlim([xmin, xmax])
    ax.set_ylim([ymin, ymax])

def histogram_2d(
                 ax,
                 x,
                 y,
                 w = None,
                 xmin = None,
                 xmax = None,
                 ymin = None,
                 ymax = None,
                 log_scale = False,
                 log_offset = 1e-4,
                 cmap = 'terrain_r',
                 bins = 100,
                ):
    '''\
    Quickly make a histogram of some samples
    '''
    import numpy as np
    
    # Update limits
    if xmin is None:
        xmin = np.min(x)
    if xmax is None:
        xmax = np.max(x)
    if ymin is None:
        ymin = np.min(y)
    if ymax is None:
        ymax = np.max(y)

    # Make histogram
    rho, xedge, yedge = \
            np.histogram2d(
                           x, y,
                           weights = w,
                           range = [[xmin, xmax], [ymin, ymax]],
                           bins = bins,
                          )

    # log scale
    if log_scale:
        # Check min val
        min_val = np.min(rho[rho > 0.])
        # Implement log scale
        rho = np.log(rho + log_offset*min_val)

    # Show data
    im = ax.imshow(
              np.rot90(rho),
              extent = [xmin, xmax, ymin, ymax],
              aspect = 'auto',
              cmap = cmap,
             )

    ax.set_xlim([xmin, xmax])
    ax.set_ylim([ymin, ymax])
    return im
                     

def density_contour_2d(
                       ax,
                       x,
                       y,
                       w = None,
                       xmin = None,
                       xmax = None,
                       ymin = None,
                       ymax = None,
                        cmap = 'magma',
                       bins = 100,
                       levels = [0.25, 0.5, 0.75],
                       **kwargs
                      ):
    '''\
    Plot density contours
    '''
    import numpy as np
    from scipy.stats import gaussian_kde
    
    # Update limits
    if xmin is None:
        xmin = np.min(x)
    if xmax is None:
        xmax = np.max(x)
    if ymin is None:
        ymin = np.min(y)
    if ymax is None:
        ymax = np.max(y)

    # Generate inputs
    xspace = np.linspace(xmin, xmax, bins)
    yspace = np.linspace(ymin, ymax, bins)
    xgrid, ygrid = np.meshgrid(xspace, yspace)

    # Stack inputs
    rv_input = np.stack((xgrid.flatten(), ygrid.flatten()))

    # fit kde
    K = gaussian_kde(np.asarray([x,y]), weights = w)
    rho = K.pdf(rv_input).reshape((bins, bins)).T

    # Get values
    #rho, xedge, yedge = \
    #        np.histogram2d(
    #                       x, y,
    #                       weights = w,
    #                       range = [[xmin, xmax], [ymin, ymax]],
    #                       bins = bins,
    #                      )
    rho[~np.isfinite(rho)] = 0.0
    rho = rho /np.sum(rho)

    # Generate contour levels
    levels = percentile_levels(xgrid, ygrid, rho, levels)

    # Plot contours
    ax.contour(
               xspace,
               yspace,
               rho.T,
               levels = levels,
               cmap = cmap,
               vmin = 0.,
               vmax = np.max(rho),
               **kwargs
              )
    
    # Set limits
    ax.set_xlim([xmin, xmax])
    ax.set_ylim([ymin, ymax])
                    
def gaussian_pdf_2d(
                    ax,
                    x,
                    y,
                    w = None,
                    xmin = None,
                    xmax = None,
                    ymin = None,
                    ymax = None,
                    log_scale = False,
                    log_offset = 1e-4,
                    cmap = 'terrain_r',
                    bins = 100,
                    approx = None,
                   ):
    '''\
    Quickly make a histogram of some samples
    '''
    import numpy as np
    # Complain about Gaussian
    if approx is None:
        raise Exception("Gaussian parameters unknown to plotting function")
    
    # Update limits
    if xmin is None:
        xmin = np.min(x)
    if xmax is None:
        xmax = np.max(x)
    if ymin is None:
        ymin = np.min(y)
    if ymax is None:
        ymax = np.max(y)

    # Reorganize samples
    samples = [x, y]
    limits = [[xmin, xmax], [ymin, ymax]]

    # Generate inputs
    xspace = np.linspace(xmin, xmax, bins)
    yspace = np.linspace(ymin, ymax, bins)
    xgrid, ygrid = np.meshgrid(xspace, yspace)

    # Stack inputs
    rv_input = np.stack((xgrid, ygrid)).T
    
    # Get values
    pdf = approx.pdf(rv_input)
    pdf[~np.isfinite(pdf)] = 0.0

    # log scale
    if log_scale:
        # Check min val
        min_val = np.min(pdf[pdf > 0.])
        #max_val = np.max(pdf[pdf > 0.])
        # Implement log scale
        pdf = np.log(pdf + log_offset*min_val)

    # Show data
    ax.imshow(
              np.rot90(pdf),
              extent = [xmin, xmax, ymin, ymax],
              aspect = 'auto',
              cmap = cmap,
             )
    

    ax.set_xlim([xmin, xmax])
    ax.set_ylim([ymin, ymax])
                     



#!/usr/bin/env python3
'''\
Test the Mesh object fitting, saving, and loading
'''
######## IMPORTS ########
import numpy as np

######## Settings ########

N_RVS = int(1e5)
NSELECT = 1000
NDIM = 3
NDRAWS = 10
HYPERCUBE_RES = 100
SEED = 0
RS = np.random.RandomState(SEED)
LIMITS = np.asarray([
                     [-7., 7.],
                     [-7., 7.],
                     [-7., 7.],
                    ])


######## Tests ########
def test_posterior():
    #### Imports ####
    import time
    import numpy as np
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.optimize_likelihood_grid import optimize_likelihood_grid
    from gwalk.multivariate_normal import pdf as multivariate_normal_pdf
    from gwalk.multivariate_normal import params_of_offset_mu_cov
    from gwalk.multivariate_normal import mu_cov_of_params
    from gwalk.multivariate_normal import cor_of_cov
    from gwalk.multivariate_normal import offset_of_params
    from gwalk.multivariate_normal import MultivariateNormal
    from scipy.stats import multivariate_normal
    from figaro.mixture import DPGMM
    from figaro.load import load_density, save_density
    from tqdm import tqdm

    #### Generate a grid ####
    RS.seed(SEED)
    X0 = random_gauss_params(1, NDIM, rs=RS)
    mu, cov = mu_cov_of_params(X0)
    X0 = X0.flatten()
    #objective = "kl"
    objective = "lstsq"
    method="SLSQP"

    #### run code ####

    ## random samples ##
    #samples = (RS.uniform(size=(N_RVS,NDIM))*(LIMITS[:,1]-LIMITS[:,0])) + LIMITS[:,0]
    samples = multivariate_normal.rvs(mean=mu[0], cov=cov[0], size=N_RVS, random_state=RS)
    for i in range(NDIM):
        mask = (samples[:,i] > LIMITS[i,0]) & (samples[:,i] < LIMITS[i, 1])
        samples = samples[mask]

    t0 = time.time()
    # Create a DPGMM
    mixture = DPGMM(LIMITS)
    # Add samples to DPGMM
    for s in samples:
        mixture.add_new_point(s)
    # realize the DPGMM
    realization = mixture.build_mixture()
    # initialize the mixture
    mixture.initialise()

    # get some samples
    mix_samples = realization.rvs(NSELECT)

    # Get some draws
    draws = []
    for _ in tqdm(range(NDRAWS)):
        draws.append(mixture.density_from_samples(samples))

    # save them
    save_density(draws, "./data", name="DPGMM_draws", ext="pkl")
    # Load them
    draws = load_density("./data/DPGMM_draws.pkl")

    # Get the logpdf
    draw_logpdfs = np.array([d.logpdf(mix_samples) for d in draws])
    mix_logpdf = np.median(draw_logpdfs,axis=0)

    ## Generate from samples ##
    MVsimple = MultivariateNormal.from_samples(mix_samples, seed=SEED)

    ## TODO KEEP ##

    ## Fit grid ##
    t1 = time.time()
    MV = optimize_likelihood_grid(
                                  mix_samples,
                                  lnP=mix_logpdf,
                                  seed=SEED,
                                  objective=objective,
                                  method=method,
                                 )
    t2 = time.time()


    Xg = MV.params
    ## Print parameter values ##
    print("DPGMM time: %f seconds"%(t1-t0))
    print("fit time: %f seconds"%(t2-t1))
    print("Objective: %s"%(objective))
    print("--------")
    print("True params:")
    print(X0)
    #TODO
    #print("kl divergence with true params:")
    #print(grid.nal_kl_div(MV,params))
    print("--------")
    print("Simple params:")
    print(MVsimple.params)
    #print("kl divergence with simple params:")
    #print(grid.nal_kl_div(MV,grid.X_simple))
    print("ln(simple error:)")
    print(np.log(np.abs(MVsimple.params - X0)))
    print("--------")
    print("Optimized params")
    print(Xg)
    #print("kl divergence with optimized params:")
    #print(grid.nal_kl_div(MV,Xg))
    print("ln(optimized error:)")
    print(np.log(np.abs(Xg - X0)))
    return

######## Main ########
def main():
    test_posterior()
    return

######## Execution ########
if __name__ == "__main__":
    main()



#! /usr/env/bin python3
'''\
The primary scipt for making full use of GWALK
'''
######## Imports ########

import numpy as np
from gwalk.density.grid import Grid

######## Functions ########

def fit_grid(
             sample,
             values,
             limits,
             seed=0,
             p_labels=None,
             carryover=0.03,
             sig_factor=3.,
             objective="lstsq",
             method="L-BFGS-B",
             norm_limit=None,
             sig_max=None,
            ):
    ''' Fit multivariate normal object to samples with values

    Parameters
    ----------
    sample: array like, shape = (npts, ndim)
        Input sample locations
    values: array like, shape = (npts,)
        Input sample values
    limits: array like, shape = (ndim, 2)
        Input bounded interval limits
    seed: int, optional
        Input seed for random numbers
    nwalk: int, optional
        Input number of random walkers
    nstep: int, optional
        Input number of walker steps
    carryover: float, optional
        Input fraction of best guesses carried over through genetic algorithm
    sig_factor: float, optional
        Input controls jump sizes
    '''
    # Check norm limit
    if norm_limit is None:
        norm_limit = (np.max(values) - np.min(values)) + np.mean(values)
    # Initialize grid
    grid = Grid(sample, values, limits, norm_limit=norm_limit)
    # Generate a multivariate_normal object
    MV = grid.construct_nal(seed=seed,labels=p_labels,sig_max=sig_max)
    # Do simple fit
    grid.nal_fit_to_samples(MV)
    # Do a scipy optimize
    grid.nal_fit_scipy(MV,objective=objective,method=method)
    return grid, MV



'''Python evaluation of C extensions for multivariate normal distributions'''

######## Imports ########
import numpy as np
from ._decomposition import _cor_of_params
from ._decomposition import _cov_of_params
from ._decomposition import _cov_of_std_cor
from ._decomposition import _std_of_cov
from ._decomposition import _cor_of_std_cov
from ._decomposition import _params_of_offset_mu_std_cor

######## Declarations ########

__all__ = [
           "nparams_of_ndim",
           "ndim_of_nparams",
           "offset_of_params",
           "mu_of_params",
           "std_of_params",
           "cor_of_params",
           "cov_of_params",
           "mu_cov_of_params",
           "offset_mu_cov_of_params",
           "cov_of_std_cor",
           "std_of_cov",
           "cor_of_std_cov",
           "cor_of_cov",
           "params_of_offset_mu_std_cor",
           "params_of_offset_mu_cov",
           "params_of_mu_cov",
           "params_reduce",
           "params_rescale",
           "cov_rescale",
          ]

######## Functions ########
#### General functions ####
## dimension checking/reduction ##
def nparams_of_ndim(ndim):
    '''Return the number of parameters associated witha guess in n dimensions
    Parameters
    ----------
    ndim: int
        Input number of dimensions
    '''
    return int((ndim*ndim + 3*ndim)//2) + 1

def ndim_of_nparams(nparam):
    '''find the dimensionality of a guess
    Parameters
    ----------
    nparam: int
        Input number of parameters
    Returns
    -------
    ndim: int
        Number of dimensions
    '''
    for i in range(nparam):
        if nparam == nparams_of_ndim(i):
            return i
    raise RuntimeError("Failed to get number of dimensions for guess")

def offset_of_params(X):
    '''Reconstruct the normalization constant from parameters
    Parameters
    ----------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    Returns
    -------
    offset: `~numpy.ndarray` shape = (ngauss,)
        Output array of parameter offsets
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    offset = X[:,0].copy()
    return offset

def mu_of_params(X):
    '''Reconstruct the vectorized mu parameters of input Gaussian parameters

    Parameters
    ----------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    Returns
    -------
    mu: `~numpy.ndarray` shape = (ngauss, ndim)
        Output array of mu values
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    # Identify ngauss
    ngauss = X.shape[0]
    # Identify ndim
    nparams = X.shape[1]
    ndim = ndim_of_nparams(nparams)
    # Reconstruct mu
    mu = np.copy(X[:,1:ndim+1])
    return mu
    
def std_of_params(X):
    '''Reconstruct the vectorized sigma parameters of input Gaussian parameters
    Parameters
    ----------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    Returns
    -------
    std: `~numpy.ndarray` shape = (ngauss, ndim)
        Output array of std values
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    # Identify ngauss
    ngauss = X.shape[0]
    # Identify ndim
    nparam = X.shape[1]
    ndim = ndim_of_nparams(nparam)
    # Reconstruct std
    std = np.copy(X[:,ndim+1:2*ndim+1])
    return std

def cor_of_params(X):
    '''Reconstruct the vectorized correlation matrix of input Gaussian parameters
    Parameters
    ----------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    Returns
    -------
    cor: `~numpy.ndarray` shape = (ngauss, ndim, ndim)
        Output array of correlation matrices
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    # Identify ngauss
    ngauss = X.shape[0]
    # Identify ndim
    nparam = X.shape[1]
    ndim = ndim_of_nparams(nparam)
    # Reconstruct correlation matrices
    cor = _cor_of_params(X, ndim)
    return cor

def cov_of_params(X):
    '''Reconstruct the vectorized corvariance of input Gaussian parameters
    Parameters
    ----------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    Returns
    -------
    cov: `~numpy.ndarray` shape = (ngauss, ndim, ndim)
        Output array of covariance matrices
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    # Identify ngauss
    ngauss = X.shape[0]
    # Identify ndim
    nparam = X.shape[1]
    ndim = ndim_of_nparams(nparam)
    # Get cov
    cov = _cov_of_params(X, ndim)
    return cov

def mu_cov_of_params(X):
    '''Reconstruct vectorized mu and covariance of input Gaussian parameters
    Parameters
    ----------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    Returns
    -------
    mu: `~numpy.ndarray` shape = (ngauss, ndim)
        Output array of mu values
    cov: `~numpy.ndarray` shape = (ngauss, ndim, ndim)
        Output array of covariance matrices
    '''
    mu = mu_of_params(X)
    cov = cov_of_params(X)
    return mu, cov

def offset_mu_cov_of_params(X):
    '''Reconstruct vectorized mu and covariance of input Gaussian parameters
    Parameters
    ----------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    Returns
    -------
    offset: `~numpy.ndarray` shape = (ngauss,)
        Output array of parameter offsets
    mu: `~numpy.ndarray` shape = (ngauss, ndim)
        Output array of mu values
    cov: `~numpy.ndarray` shape = (ngauss, ndim, ndim)
        Output array of covariance matrices
    '''
    mu = mu_of_params(X)
    cov = cov_of_params(X)
    offset = offset_of_params(X)
    return offset, mu, cov


def cov_of_std_cor(std,cor):
    '''Reconstruct the covariance from std and corelation
    Parameters
    ----------
    std: array like, shape = (ngauss, ndim)
        Input array of sigma values
    cor: array like, shape = (ngauss, ndim, ndim)
        Input array of correlation matrix values
    Returns
    -------
    cov: `~numpy.ndarray` shape = (ngauss, ndim, ndim)
        Output array of covariance matrices
    '''
    # Check dimensionality of std
    if len(std.shape) == 1:
        ngauss, ndim = 1, std.size
        std = std.reshape((ngauss,ndim))
    else:
        ngauss, ndim = std.shape
    # Check dimensionality of cor
    if len(cor.shape) == 2:
        cor = cor.reshape((ngauss,ndim,ndim))
    # Assert dimensionality match up
    assert cor.shape == ((ngauss,ndim,ndim))
    # Reconstruct the covariance matrix
    return _cov_of_std_cor(std, cor)

def std_of_cov(cov):
    '''Extract the vectorized std from the vectorized covariance
    Parameters
    ----------
    cov: array like, shape = (ngauss, ndim, ndim)
        Input Array of cov values
    Returns
    -------
    std: array like, shape = (ngauss, ndim)
        Input array of sigma values
    '''
    # Protect against usage
    if len(cov.shape) == 2:
        cov = cov[None,:,:]
    elif len(cov.shape) != 3:
        raise ValueError("Error cor_of_cov")
    # Check squareness
    ndim = cov.shape[-1]
    if not (cov.shape[-2] == ndim):
        raise ValueError("cov must be square")
    # Identify std values
    return _std_of_cov(cov)

def cor_of_std_cov(std, cov):
    '''Reconstruct the vectorized correlation matrix from the covariance
    Parameters
    ----------
    std: array like, shape = (ngauss, ndim)
        Input array of sigma values
    cov: array like, shape = (ngauss, ndim, ndim)
        Input Array of cov values
    Returns
    -------
    cor: array like, shape = (ngauss, ndim, ndim)
        Input array of correlation matrix values
    '''
    # Protect against usage
    # Check dimensionality of std
    if len(std.shape) == 1:
        ngauss, ndim = 1, std.size
        std = std.reshape((ngauss,ndim))
    else:
        ngauss, ndim = std.shape
    # Check dimensionality of cov
    if len(cov.shape) == 2:
        cov = cov[None,:,:]
    elif len(cov.shape) != 3:
        raise ValueError("Error cor_of_cov")
    # Check squareness
    ndim = cov.shape[-1]
    if not (cov.shape[-2] == ndim):
        raise ValueError("cov must be square")
    # Identify covariance
    return _cor_of_std_cov(std, cov)

def cor_of_cov(cov):
    '''Reconstruct the vectorized correlation matrix from the covariance
    Parameters
    ----------
    cov: array like, shape = (ngauss, ndim, ndim)
        Input Array of cov values
    Returns
    -------
    cor: array like, shape = (ngauss, ndim, ndim)
        Input array of correlation matrix values
    '''
    # Protect against usage
    # Check dimensionality of cov
    if len(cov.shape) == 2:
        cov = cov[None,:,:]
    elif len(cov.shape) != 3:
        raise ValueError("Error cor_of_cov")
    # Check squareness
    ndim = cov.shape[-1]
    if not (cov.shape[-2] == ndim):
        raise ValueError("cov must be square")
    # Identify std values
    std = _std_of_cov(cov)
    # Identify covariance
    return _cor_of_std_cov(std, cov)

def params_of_offset_mu_std_cor(offset, mu, std, cor):
    '''Reconstruct Gaussian vectorized parameters for given mu and covariance
    Parameters
    ----------
    offset: `~numpy.ndarray` shape = (ngauss,)
        Output array of parameter offsets
    mu: array like, shape = (ngauss, ndim)
        Input Array of mu values
    std: array like, shape = (ngauss, ndim)
        Input array of sigma values
    cor: array like, shape = (ngauss, ndim, ndim)
        Input array of correlation matrix values
    Returns
    -------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    '''
    # Protect against usage
    # Check dimensionality of mu
    if len(mu.shape) == 1:
        ngauss, ndim = 1, mu.size
        mu = mu.reshape((ngauss,ndim))
    elif(len(mu.shape) == 2):
        ngauss, ndim = mu.shape
    else:
        raise RuntimeError("Mu is not the right shape")

    # Check dimensionality of std
    if len(std.shape) == 1:
        ngauss, ndim = 1, std.size
        std = std.reshape((ngauss,ndim))
    elif (len(mu.shape) == 2):
        ngauss, ndim = std.shape
    else:
        raise RuntimeError("std is not the right shape")

    # Check dimensionality of cor
    if len(cor.shape) == 2:
        cor = cor[None,:,:]
    elif len(cor.shape) != 3:
        raise ValueError("Error cor is not the right shape")

    # Check offset
    offset = np.asarray([offset]).reshape((ngauss,))
    if not len(offset.shape) >= 1:
        raise RuntimeError("Offset is not the right shape")

    # Identify useful information
    ngauss = mu.shape[0]
    ndim = mu.shape[1]
    nparams = nparams_of_ndim(ndim)
    # Protect against usage
    if cor.shape != (ngauss, ndim, ndim):
        raise ValueError("cor is not shape (ngauss, ndim, ndim)")

    return _params_of_offset_mu_std_cor(offset, mu, std, cor, nparams)

def params_of_offset_mu_cov(offset, mu, cov):
    '''Reconstruct Gaussian vectorized parameters for given mu and covariance
    Parameters
    ----------
    offset: `~numpy.ndarray` shape = (ngauss,)
        Output array of parameter offsets
    mu: array like, shape = (ngauss, ndim)
        Input Array of mu values
    cov: array like, shape = (ngauss, ndim, ndim)
        Input array of covariance matrix values
    Returns
    -------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    '''
    # Protect against usage

    # Check dimensionality of mu
    if len(mu.shape) == 1:
        ngauss, ndim = 1, mu.size
        mu = mu.reshape((ngauss,ndim))
    elif(len(mu.shape) == 2):
        ngauss, ndim = mu.shape
    else:
        raise RuntimeError("Mu is not the right shape")

    # Check dimensionality of cov
    if len(cov.shape) == 2:
        cov = cov[None,:,:]
    elif len(cov.shape) != 3:
        raise ValueError("Error cov is not the right shape")

    # Check offset
    offset = np.asarray([offset]).reshape((ngauss,))
    if not len(offset.shape) >= 1:
        raise RuntimeError("Offset is not the right shape")

    # Identify useful information
    ngauss = mu.shape[0]
    ndim = mu.shape[1]
    nparams = nparams_of_ndim(ndim)
    # Protect against usage
    if cov.shape != (ngauss, ndim, ndim):
        raise ValueError("cor is not shape (ngauss, ndim, ndim)")

    # Identify std values
    std = _std_of_cov(cov)
    # Identify correlations
    cor = _cor_of_std_cov(std, cov)

    params = _params_of_offset_mu_std_cor(offset, mu, std, cor, nparams)
    return params

def params_of_mu_cov(mu, cov):
    '''Reconstruct Gaussian vectorized parameters for given mu and covariance
    This function assumes there are no log offsets for the pdf.
    
    Parameters
    ----------
    mu: array like, shape = (ngauss, ndim)
        Input Array of mu values
    cov: array like, shape = (ngauss, ndim, ndim)
        Input array of covariance matrix values
    Returns
    -------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    '''
    # Protect against usage
    # Check dimensionality of mu
    if len(mu.shape) == 1:
        ngauss, ndim = 1, mu.size
        mu = mu.reshape((ngauss,ndim))
    elif(len(mu.shape) == 2):
        ngauss, ndim = mu.shape
    else:
        raise RuntimeError("Mu is not the right shape")

    # Check dimensionality of cov
    if len(cov.shape) == 2:
        cov = cov[None,:,:]
    elif len(cov.shape) != 3:
        raise ValueError("Error cov is not the right shape")

    # Identify useful information
    ngauss = mu.shape[0]
    ndim = mu.shape[1]
    nparams = nparams_of_ndim(ndim)
    # Protect against usage
    if cov.shape != (ngauss, ndim, ndim):
        raise ValueError("cor is not shape (ngauss, ndim, ndim)")

    # Identify std values
    std = _std_of_cov(cov)
    # Identify correlations
    cor = _cor_of_std_cov(std, cov)

    return _params_of_offset_mu_std_cor(np.zeros(ngauss), mu, std, cor, nparams)

def params_reduce(X, indices):
    '''Reduce to indexed marginals
    Parameters
    ----------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    indices: list
        Input list of indices we would like parameters for
    Returns
    -------
    X: array like, shape = (ngauss, nparams_new)
        Output  array of parameter guesses for subset of parameters
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    # Extract dimensionality
    ngauss, nparam = X.shape
    ndim = ndim_of_nparams(nparam)
    # Extract dimensionality of indices
    ndim_index = len(indices)
    nparam_index = nparams_of_ndim(ndim_index)
    # Initialize array
    Xp = np.empty((ngauss, nparam_index))
    # Set normalization
    Xp[:,0] = X[:,0]
    # Loop through Xp dimensions
    for i in range(ndim_index):
         Xp[:,i+1] =              X[:,indices[i] + 1]
         Xp[:,i+ndim_index+1] =   X[:,indices[i]+ndim + 1]
    # Loop through correlation factors
    for i in range(ndim_index):
        for j in range(i):
            if indices[i] > indices[j]:
                index, jndex = indices[i], indices[j]
            elif indices[i] < indices[j]:
                index, jndex = indices[j], indices[i]
            else:
                raise RuntimeError("indices[i] == indices[j]")
            Xp[:,2*ndim_index + ((i*(i-1))//2) + j + 1] = \
                X[:,2*ndim + ((index*(index-1))//2) + jndex + 1]
    # flatten
    #if ngauss == 1:
    #    Xp = Xp.reshape((nparam_index,))
    return Xp

def params_rescale(X, scale):
    ''' Rescale parameters
    Parameters
    ----------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    scale: array like, shape = (ngauss, ndim)
        Input array of scale factors
    Returns
    -------
    X: array like, shape = (ngauss, nparams)
        Rescaled X matrix
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    # Extract dimensionality
    ngauss, nparam = X.shape
    ndim = ndim_of_nparams(nparam)
    # Check scale
    if len(scale.shape) == 1:
        if ndim != scale.size:
            raise RuntimeError("Invalid scale shape")
        else:
            scale = np.ones((ngauss, ndim))*scale
    else:
        assert scale.shape == ((ngauss, ndim))
    
    ## Rescale array ##
    X = X.copy()
    X[:,1:ndim+1] *= scale
    X[:,ndim+1:2*ndim+1] *= scale
    return X

def cov_rescale(cov, scale=None):
    ''' Rescale parameters
    Parameters
    ----------
    cov: `~numpy.ndarray` shape = (ngauss, ndim, ndim)
        Output array of covariance matrices
    scale: array like, shape = (ngauss, ndim)
        Input array of scale factors
    Returns
    -------
    cov: `~numpy.ndarray` shape = (ngauss, ndim, ndim)
        Rescaled X matrix
    '''
    # Simple case
    if scale is None:
        return cov.copy()
    # Protect against usage
    if len(cov.shape) == 2:
        cov = cov[None,:,:]
    elif len(cov.shape) != 3:
        raise ValueError("Error cor_of_cov")
    # Check squareness
    ndim = cov.shape[-1]
    if not (cov.shape[-2] == ndim):
        raise ValueError("cov must be square")
    # Extract dimensionality
    ngauss = cov.shape[0]
    ndim = cov.shape[1]
    # Check scale
    if len(scale.shape) == 1:
        if ndim != scale.size:
            raise RuntimeError("Invalid scale shape")
        else:
            scale = np.ones((ngauss, ndim))*scale
    else:
        assert scale.shape == ((ngauss, ndim))
    
    # Identify std values
    std = _std_of_cov(cov)
    # Identify correlations
    cor = _cor_of_std_cov(std, cov)
    # Rescale std
    std *= scale
    # Get cov back
    cov = _cov_of_std_cor(std, cor)
    return cov



#!/usr/bin/env python3
'''\
Test the Multivariate Normal object and methods
'''
######## Imports ########
from contextlib import contextmanager
from gwalk import MultivariateNormal
from gwalk.catalog import Catalog
import sys
from os.path import join, isfile, isdir
import os
import time
import numpy as np
from gwalk.catalog.coordinates import coord_tags as COORD_TAGS
from gwalk.tools.fit_catalog_samples import fit_real_event
from gwalk.tools.plot_mesh_norm_corner import plot_GW

######## Context Manager ########

@contextmanager
def custom_redirection(wkdir, logfile):
    # Check the run directory
    assert isdir(wkdir)
    # break up the logfile
    dirs = logfile.split('/')
    # Create directories
    for i in range(len(dirs)):
        if not isdir(join(wkdir,*dirs[:i])):
            os.mkdir(join(wkdir,*dirs[:i]))
    # Begin managing the context
    with open(join(wkdir,logfile),'w') as out:
        old_out = sys.stdout
        sys.stdout = out
        old_err = sys.stderr
        sys.stderr = out
        try:
            yield out
        finally:
            sys.stdout = old_out
            sys.stderr = old_err


######## Settings ########

#nal_methods = ["simple", "genetic", "select"]
nal_methods=["genetic"]
extensions = ["png", "pdf"]

# These ones are really important
convergence = "mesh_kl"
nwalk = 100
nstep = 100
evaluation_res = 100
kl_sensitivity = 1e-3
max_bins1d = 100
max_bins2d = 31
min_bins = 30

# These ones are less important
carryover = 0.05
sleep = 0

# Initialize hyperparameters
hyperparameters = {
                   "nwalk"          : nwalk,
                   "nstep"          : nstep,
                   "kl_sensitivity" : kl_sensitivity,
                   "carryover"      : carryover,
                   "convergence"    : convergence,
                  }

######## Arguments ########

wkdir = sys.argv[1]
fname_catalog = join(wkdir, "catalog_samples.hdf5")
assert isfile(fname_catalog)
fname_mesh = join(wkdir, "likelihood_mesh.hdf5")

event = sys.argv[2]
cata = Catalog(fname_catalog)
release = cata.release_of_event(event)
available_coord_tags = cata.tags_of_event(event)
fname_release = join(wkdir, "%s.nal.hdf5"%release)

######## Pipeline ########

# Delay start to reduce database lookup overlaps
sleeptime = sleep*np.random.uniform()
time.sleep(sleeptime)


# Loop the coord tags
for coord_tag in available_coord_tags:
    # Find available groups for this coord tag
    groups = cata.groups_of_event_tag(event, coord_tag)
    # Loop the groups
    for group in groups:
        # Check for singular values
        singular = False
        sample_dict = cata.load_data(event,group,COORD_TAGS[coord_tag])
        for key in sample_dict:
            if min(sample_dict[key]) == max(sample_dict[key]):
                singular = True
        if singular:
            continue
        # Update the user
        with open("nohup.out", 'a') as File:
            print(event, coord_tag, group, file=File)
        # Loop the fit methods
        for fit_method in nal_methods:
            # Identify fit log file
            fit_log = join("logs", event, "%s:%s:%s_fit.log"%(
                coord_tag, group, fit_method))
            # Open context manager
            with custom_redirection(wkdir, fit_log):
                fit_real_event(
                               fname_catalog,
                               fname_release,
                               event,
                               group,
                               coord_tag,
                               min_bins=min_bins,
                               max_bins1d=max_bins1d,
                               max_bins2d=max_bins2d,
                               evaluation_res=evaluation_res,
                               nwalk=nwalk,
                               nstep=nstep,
                               whitenoise=0.001,
                               sig_factor=0.5,
                               carryover=carryover,
                               fname_mesh=fname_mesh,
                               verbose=True,
                              )

            # Identify plot log file
            plot_log = join("logs", event, "%s:%s:%s_plot.log"%(
                coord_tag, group, fit_method))
            # Make sure there is a figures folder
            if not isdir(join(wkdir,"figures")):
                os.mkdir(join(wkdir,"figures"))
            # Make sure there is an event figures folder
            if not isdir(join(wkdir,"figures",event)):
                os.mkdir(join(wkdir,"figures",event))
            # Open context manager
            with custom_redirection(wkdir, plot_log):

                # loop extensions
                for extension in extensions:
                    # Identify plot name
                    fname_plot = join(wkdir, "figures", event, "%s:%s:%s:%s_likelihood.%s"%(
                        event, coord_tag, group, fit_method, extension))
                    # Identify labels
                    label_mesh = "%s/%s:%s"%(event,coord_tag,group)
                    label_nal = "%s/%s:%s:%s"%(event,coord_tag,group,fit_method)
                    # Generate plot
                    plot_GW(
                            fname_mesh,
                            fname_release,
                            fname_plot,
                            event,
                            coord_tag,
                            group,
                            fit_method,
                            evaluation_res=evaluation_res,
                           )




'''\
Read samples from GWTC-1_sample_release and convert to source frame
'''

######## Imports ########
from gwalk.catalog.coordinates import m1m2_from_mc_eta, mc_eta_from_m1m2
from gwalk.catalog.coordinates import z_from_lum_dist_interp
from gwalk.catalog.coordinates import source_from_detector
from gwalk.catalog.coordinates import detector_from_source
from gwalk.catalog.coordinates import chieff_from_m1m2s1s2
from gwalk.catalog.coordinates import coord_labels
from gwalk.catalog.coordinates import lambdatilde_of_eta_lam1_lam2
from gwalk.catalog.coordinates import deltalambda_of_eta_lam1_lam2
from os.path import join, isfile
from xdata.database import Database

######## Functions ########
#### Database handling ####

def get_samples(
                fname_event,
                names,
                group = "C01:IMRPhenomPv3HM",
               ):
    '''Get the posterior samples we need

    Parameters
    ----------
    fname_event: str
        Input file name  for catalog samples
    names: list
        Input list of names to draw from catalog samples
    group: str
        Input kind of samples to work with
    '''

    # Check if file exists
    assert isfile(fname_event)

    # Initialize database
    db = Database(fname_event)
    # Check if group exists
    assert db.exists(group)

    # check if data exists
    data_addr = group
    assert db.exists(data_addr, kind="dset")

    # Get the fields
    fields = db.dset_fields(data_addr)

    # Create a dictionary
    pdict = {}
    
    # Load each item
    for item in names:
        # Make sure the item is correct
        if item in fields:
            # append values
            pdict[item] = db.dset_value(data_addr, field=item)

    return pdict

#### Calculations ####

def generate_samples(
                     fname_event,
                     group,
                     z_bins = 100,
                    ):
    '''Generate samples from GWTC-1 catalog release
    Parameters
    ----------
    fname_event: str
        Input file name  for catalog samples
    group: str
        Input kind of samples to work with
    z_bins: int
        Input number of bins for redshift interpolation
    '''
    import numpy as np

    names = [
             "costheta_jn",
             "luminosity_distance_Mpc",
             "right_ascension",
             "declination",
             "m1_detector_frame_Msun",
             "m2_detector_frame_Msun",
             "spin1",
             "spin2",
             "costilt1",
             "costilt2",
             "lambda1",
             "lambda2",
            ]

    #### Algorithm ####
    # Generate samples
    name_dict = \
            get_samples(
                        fname_event,
                        names,
                        group=group,
                       )

    pdict = {}
    pdict["cos_tilt_1"] = name_dict["costilt1"]
    pdict["cos_tilt_2"] = name_dict["costilt2"]

    ## Find redshift
    # Load luminosity distance in Mpc
    lum_dist = np.copy(name_dict["luminosity_distance_Mpc"])
    # Find redshift
    Z = z_from_lum_dist_interp(lum_dist, nbins = z_bins)
    # Append to dictionary
    pdict["redshift"] = np.copy(Z)



    pdict["cos_theta_jn"] = name_dict["costheta_jn"]
    pdict["luminosity_distance"] = name_dict["luminosity_distance_Mpc"]
    pdict["ra"] = name_dict["right_ascension"]
    pdict["dec"] = name_dict["declination"]
    pdict["mass_1"] = name_dict["m1_detector_frame_Msun"]
    pdict["mass_2"] = name_dict["m2_detector_frame_Msun"]
    pdict["a_1"] = name_dict["spin1"]
    pdict["a_2"] = name_dict["spin2"]

    # Find source frame masses
    mc, eta = mc_eta_from_m1m2(name_dict["m1_detector_frame_Msun"],
                               name_dict["m2_detector_frame_Msun"])
    m1_source = source_from_detector(name_dict["m1_detector_frame_Msun"], Z)
    m2_source = source_from_detector(name_dict["m2_detector_frame_Msun"], Z)
    mc_source, eta_source = mc_eta_from_m1m2(m1_source, m2_source)
    # Find Chieff
    spin_1z = pdict["a_1"]*pdict["cos_tilt_1"]
    spin_2z = pdict["a_2"]*pdict["cos_tilt_2"]
    spin_1xy = pdict["a_1"]*np.power(1 - pdict["cos_tilt_1"]**2, 0.5)
    spin_2xy = pdict["a_2"]*np.power(1 - pdict["cos_tilt_2"]**2, 0.5)
    chieff = chieff_from_m1m2s1s2(m1_source, m2_source, spin_1z, spin_2z)
    # Append to dictionary
    pdict["mass_1_source"] = m1_source
    pdict["mass_2_source"] = m2_source
    pdict["chirp_mass_source"] = mc_source
    pdict["chirp_mass"] = mc
    pdict["symmetric_mass_ratio"] = eta
    pdict["chi_eff"] = chieff
    pdict["spin_1z"] = spin_1z
    pdict["spin_2z"] = spin_2z
    pdict["spin_1xy"] = spin_1xy
    pdict["spin_2xy"] = spin_2xy
    pdict["inv_lum_dist"] = np.power(pdict["luminosity_distance"],-1.)
    pdict["total_mass"] = pdict["mass_1"] + pdict["mass_2"]
    pdict["total_mass_source"] = pdict["mass_1_source"] + pdict["mass_2_source"]
    pdict["mass_ratio"] = pdict["mass_2"]/pdict["mass_1"]

    # Try to add lambda parameters
    try:
        pdict["lambda_1"] = name_dict["lambda1"]
        pdict["lambda_2"] = name_dict["lambda2"]
        pdict["lambda_tilde"] = lambdatilde_of_eta_lam1_lam2(
                                    pdict["symmetric_mass_ratio"],
                                    pdict["lambda_1"],
                                    pdict["lambda_2"],
                                   )
        pdict["delta_lambda"] = deltalambda_of_eta_lam1_lam2(
                                    pdict["symmetric_mass_ratio"],
                                    pdict["lambda_1"],
                                    pdict["lambda_2"],
                                   )
    except:
        pass

    # Return dictionary
    return pdict



#!/home/xevra/.local/bin/python3
'''\
Histogram tools
'''
######## Functions ########

def get_bin_ids(X, limits, nbins):
    '''return each bin id for a 1D histogram in each dimension

    Parameters
    ----------
    X: array like, shape = (npts, ndim)
        Input values for histogram
    limits: array like, shape = (2, ndim)
        Input limits for histogram
    nbins: int
        Input bins for histogram
    '''
    import numpy as np
    # Check dimensionality
    ndim = limits.shape[0]
    if not X.shape[1] == ndim:
        raise RuntimeError("X shape is %s, but ndim is %d"%(str(X.shape),ndim))
    # Identify edges
    edges =     np.empty((ndim,nbins+1),dtype=float)
    centers =   np.empty((ndim,nbins),dtype=float)
    for i in range(ndim):
        edges[i] = np.linspace(limits[i,0],limits[i,1],nbins+1)
        centers[i] = 0.5*(edges[i,1:] + edges[i,:-1])

    # Initialize bin ids
    bin_ids = np.empty(X.shape,dtype=int)

    # Indentify bin ids
    for i in range(ndim):
        for j in range(nbins):
            in_bin = (X[:,i] > edges[i,j]) & (X[:,i] < edges[i,j + 1])
            bin_ids[in_bin,i] = j
    return bin_ids, edges, centers

def hist1d_from_ids(bin_ids, dim1, nbins, weights=None):
    '''return a 1D histogram for from bin_ids and weights
    Parameters
    ----------
    bin_ids: array like, shape = (npts,ndim)
        Input ids for bins
    dim1: int
        Input id for dimension
    nbins: int
        Input bins for histogram
    weights: array like, shape = (npts)
        Input weights for histogram
    '''
    import numpy as np
    # Check dimensionality
    if not weights.size == bin_ids.shape[0]:
        raise RuntimeError("bin_ids shape is %s, weights shape is %s"%(\
                str(bin_ids.shape),str(weights.shape)))

    # Initialize weights
    if weights is None:
        weights = np.ones(bin_ids.shape[0])

    # Initialize histogram
    hist1d = np.zeros(nbins, dtype=float)
    # Loop through each bin
    for i in range(nbins):
        hist1d[i] = np.sum(weights * (bin_ids[:,dim1] == i).astype(int))
    return hist1d / np.sum(hist1d)

def hist2d_from_ids(bin_ids, dim1, dim2, nbins, weights=None):
    '''return a 1D histogram for from bin_ids and weights
    Parameters
    ----------
    bin_ids: array like, shape = (npts,ndim)
        Input ids for bins
    dim1: int
        Input id for dimension
    dim2: int
        Input id for dimension
    nbins: int
        Input bins for histogram
    weights: array like, shape = (npts)
        Input weights for histogram
    '''
    import numpy as np
    # Check dimensionality
    if not weights.size == bin_ids.shape[0]:
        raise RuntimeError("bin_ids shape is %s, weights shape is %s"%(\
                str(bin_ids.shape),str(weights.shape)))

    # Initialize weights
    if weights is None:
        weights = np.ones(bin_ids.shape[0])

    # Initialize histogram
    hist2d = np.zeros((nbins, nbins), dtype=float)
    # Loop through each bin
    for i in range(nbins):
        for j in range(nbins):
            hist2d[i,j] = np.sum(weights*((bin_ids[:,dim1] == i) & (bin_ids[:,dim2] == j)).astype(int))
    return hist2d / np.sum(hist2d)





def make_walker_plot(MLE_likelihood, plotdir, method, event):
    '''\
    Plot walker progress

    Inputs:
        MLE_likelihood: array of walker likelihood values
        nstep: number of steps walkers take

    '''
    ####  Imports ####
    ## Matplotlib
    import matplotlib
    matplotlib.use("Agg")
    from matplotlib import pyplot as plt
    ## Public
    from os.path import join

    plt.style.use("bmh")
    fig, ax = plt.subplots()
    for i in range(MLE_likelihood.shape[0]):
        ax.plot(np.arange(MLE_likelihood.shape[1]), MLE_likelihood[i,:], 'o',markersize=2)

    savename = join(plotdir, "NAL_convergence_%s_%s.png"%(method, event))
    plt.savefig(savename)



#!/home/xevra/.local/bin/python3
'''\
Generate the likelihood function of each event
'''

######## Imports ########
import numpy as np
from gwalk.catalog.coordinates import m1m2_from_mc_eta, mc_eta_from_m1m2
from gwalk.catalog.coordinates import z_from_lum_dist_interp
from gwalk.catalog.coordinates import source_from_detector
from gwalk.catalog.coordinates import chieff_from_m1m2s1s2
from gwalk.catalog.coordinates import coord_labels
from xdata.database import Database
from os.path import join, isfile

######## Functions ########
#### Database handling ####
def get_samples(
                fname_event,
                names,
                group = "PublicationSamples",
               ):
    '''Get the posterior samples we need

    Parameters
    ----------
    fname_event: str
        Input file name  for catalog samples
    names: list
        Input list of names to draw from catalog samples
    group: str
        Input kind of samples to work with
    '''

    # Check if file exists
    assert isfile(fname_event)

    # Initialize database
    db = Database(fname_event)
    # Check if group exists
    assert db.exists(group, kind="group")

    # check if data exists
    data_addr = join(group, "posterior_samples")
    assert db.exists(data_addr, kind="dset")

    # Get the fields
    fields = db.dset_fields(data_addr)

    # Create a dictionary
    pdict = {}
    
    # Load each item
    for item in names:
        # Make sure the item is correct
        if item in fields:
            # append values
            pdict[item] = db.dset_value(data_addr, field=item)

    return pdict

######## Algorithm ########

def generate_samples(fname_event, group, **kwargs):
    '''A wrapper for finding the sample values that we need
    Parameters
    ----------
    fname_event: str
        Input file name  for catalog samples
    group: str
        Input kind of samples to work with
    '''

    names = list(coord_labels.keys())

    pdict = \
            get_samples(
                        fname_event,
                        names,
                        group=group,
                       )

    # Generate some things not given
    pdict["inv_lum_dist"] = np.power(pdict["luminosity_distance"],-1.)

    # Check on spin coordinates
    if not "chi_eff" in pdict:
        pdict["chi_eff"] = chieff_from_m1m2s1s2(
                                                pdict["mass_1_source"],
                                                pdict["mass_2_source"],
                                                pdict["a_1"],
                                                pdict["a_2"],
                                               )
    # Check aligned spin
    if not "spin_1z" in pdict:
        pdict["spin_1z"] = pdict["a_1"]*pdict["cos_tilt_1"]
    if not "spin_2z" in pdict:
        pdict["spin_2z"] = pdict["a_2"]*pdict["cos_tilt_2"]

    # check planar spin
    pdict["spin_1xy"] = pdict["a_1"]*np.power(1 - pdict["cos_tilt_1"]**2, 0.5)
    pdict["spin_2xy"] = pdict["a_2"]*np.power(1 - pdict["cos_tilt_2"]**2, 0.5)

    return pdict
